"""Module containing hardcoded logging json configs.

References: https://docs.python.org/3/library/logging.config.html#logging-config-dictschema
"""

# Standard libs
from typing import Dict


# pylint: disable=too-few-public-methods
class AppLoggerConfig:
    """Class to define hard-coded json logging configs."""

    @staticmethod
    def default_format_1() -> Dict:
        """Define the logging configuration in json for the `default_format_1`.

        Note:
            format => name, levelname, message with a date format of `yyyy-mm-ddThh-MM-ss` in a single line.
            handlers => console output.

        :return: Configuration of the formatters, filters, handlers, and loggers as well as the root logger.
        :rtype: Returns a dictionary object.
        """
        return {
            "version": 1,  # This has to be 1
            "disable_existing_loggers": True,
            "incremental": False,
            "formatters": {
                "default-single-line": {
                    "format": "{name:<10s} | {levelname:8s} | {message:s}",
                    "datefmt": "%Y-%m-%dT%H:%M:%S",
                    "style": "{",
                }
            },
            "filters": {},
            "handlers": {
                "console": {
                    "class": "logging.StreamHandler",
                    "level": "INFO",
                    "formatter": "default-single-line",
                    "filters": "",
                    "stream": "ext://sys.stdout",
                }
            },
            "loggers": {},
            "root": {"handlers": ["console"]},
        }



"""Module containing logging config names enum."""

# Standard libs
from enum import Enum


# pylint: disable=too-few-public-methods
class AppLoggerConfigName(Enum):
    """Class to define logging config names enum value pair."""

    DEFAULT_FORMAT_1 = "default_format_1"



"""Module to format logging record information into json."""

# Standard libs
import json
import logging
import decimal


class JsonFormatter(logging.Formatter):
    """Class to format logging record into json."""

    def __init__(self, fmt: str, datefmt: str, delimiter: str) -> None:
        """Initialize the class with variables.

        :param fmt: string format
        :param datefmt: date datefmt
        :param delimiter: format delimiter
        :return: Nothing
        """
        super().__init__(fmt=fmt, datefmt=datefmt)
        self.fmt = [item.strip() for item in fmt.split(",")]
        self.datefmt = datefmt
        self.delimiter = delimiter

    def format(self, record: logging.LogRecord) -> str:
        """Create a json format of the logging record.

        :param record: logging record with logging information used to format the json message
        :return: Json string with logging information.
        """
        output = {k: str(v) for k, v in record.__dict__.items() if k in self.fmt}

        if "asctime" in self.fmt and "msecs" in record.__dict__.keys():
            msecs = round(decimal.Decimal(record.__dict__["msecs"]), 0)
            output["asctime"] = f'{output["asctime"]}.{msecs}'
        return json.dumps(obj=output, default=str)



"""Module to define custom logging filter."""

# Standard lib
import logging
import socket


# pylint: disable=too-few-public-methods
class AppLoggerFilter(logging.Filter):
    """Class of custom AppLogger filter."""

    def filter(self, record: logging.LogRecord) -> bool:
        """Set the filter information in the record.

        :param record: Logging record
        :return: True indicating the information is good.
        """
        record.client_ip = self._get_client_ip()
        record.host_name = self._get_host_name()
        record.functionName = self._get_function_name()

        return True

    def _get_client_ip(self) -> str:
        """Get the client ip where the code is running on.

        :return: The ipaddress of the current hostname
        """
        hostname = self._get_host_name()
        ip_address = socket.gethostbyname(hostname)
        return ip_address

    def _get_host_name(self) -> str:
        """Get the hostname of the application this code is running on.

        :return: Hostname that serves up this code.
        """
        return socket.gethostname()

    def _get_function_name(self) -> str:
        """Get the name of the function this code is running on.

        Note: Since we are wrapping the debug, info, warning, error, fatal, and critical methods
              of the python `logging` class, we have to go into the trace stack to get the calling
              function name one level up from the wrapper class

        TODO: Complete the finding of the calling function in the trace stack.

        :return: Name of the function that called the logger.
        """
        return "Unknown"



"""Main AppLogger Singleton Class Module."""

# Standard Libs
import logging
import logging.config
from typing import Any, Dict, Optional

# Custom libs
from zeppos_logging.app_logger_config import AppLoggerConfig
from zeppos_logging.app_logger_configs import AppLoggerConfigs
from zeppos_logging.app_logger_config_name import AppLoggerConfigName


# pylint: disable=too-few-public-methods
class AppLogger:
    """Singleton AppLogger Class."""

    logging.basicConfig(level=logging.INFO)
    __logger: logging.Logger = logging.getLogger()
    __logger.addHandler(logging.NullHandler())
    config_dict: Dict[str, Any]

    @staticmethod
    def apply_configuration(
        app_logger_config_name: Optional[AppLoggerConfigName] = None,
    ) -> None:
        """Apply the yaml configuration if any is available.

        :param app_logger_config_name: the name of the config to be used to log information
            Available configs:
            - default_format_1
            - mqtt_format_1

            By default the `default_format_1` gets set which is logging to the console only.
        :return: Nothing
        """
        app_logger_config_name_resolved: AppLoggerConfigName = (
            AppLoggerConfigName.DEFAULT_FORMAT_1 if app_logger_config_name is None else app_logger_config_name
        )
        AppLogger.debug("Applying json configuration.")
        AppLogger._configure_logger(app_logger_config_name=app_logger_config_name_resolved)

    @staticmethod
    def _configure_logger(app_logger_config_name: AppLoggerConfigName) -> None:
        """Set the logger configuration.

        :param app_logger_config_name: the name of the config to be used to log information
            Available configs:
            - default_format_1
            - mqtt_format_1

        :return: Nothing
        """
        AppLogger._set_config_dict_for_config_section(app_logger_config_name=app_logger_config_name)
        # set the retrieved config_dict on the actual python logging object
        logging.config.dictConfig(AppLogger.config_dict)

    @staticmethod
    def _set_config_dict_for_config_section(app_logger_config_name: AppLoggerConfigName) -> None:
        """Set the config dictionary.

        :param app_logger_config_name: the name of the config to be used to log information
            Available configs:
            - default_format_1
            - mqtt_format_1

        :return: Nothing
        """
        AppLogger.debug(message=f"Getting Config Section for [{app_logger_config_name.value}]")
        configs = AppLoggerConfigs()

        if app_logger_config_name in configs.keys():
            AppLogger.debug(message=f"Found `{app_logger_config_name.value}` configuration.")
            AppLogger.config_dict = configs[app_logger_config_name]()
        else:
            AppLogger.debug(message=f"`{app_logger_config_name.value}` configuration not found.")
            AppLogger.debug(message="Default fall-back logging enabled.")
            AppLogger.config_dict = AppLoggerConfig.default_format_1()

    @staticmethod
    def set_level(logging_level: int) -> None:
        """Set the logging level for the logger at the root level.

        :param logging_level: The level to set the root level logger to.
        :return: Nothing
        """
        AppLogger.__logger.setLevel(logging_level)

    @staticmethod
    def set_debug_level() -> None:
        """Set the logging level to debug.

        :return: Nothing
        """
        AppLogger.set_level(logging.DEBUG)

    @staticmethod
    def set_info_level() -> None:
        """Set the logging level to info.

        :return: Nothing
        """
        AppLogger.set_level(logging.INFO)

    @staticmethod
    def set_error_level() -> None:
        """Set the logging level to error.

        :return: Nothing
        """
        AppLogger.set_level(logging.ERROR)

    @staticmethod
    def set_warning_level() -> None:
        """Set the logging level to warning.

        :return: Nothing
        """
        AppLogger.set_level(logging.WARNING)

    @staticmethod
    def set_critical_level() -> None:
        """Set the logging level to critical.

        Note: Critical and Fatal are the same level.

        :return: Nothing
        """
        AppLogger.set_level(logging.CRITICAL)

    @staticmethod
    def set_fatal_level() -> None:
        """Set the logging level to fatal.

        Note: Critical and Fatal are the same level.

        :return: Nothing
        """
        AppLogger.set_level(logging.FATAL)

    @staticmethod
    def debug(message: str) -> None:
        """Log at debug level.

        :param message: Message to log
        :return: Nothing
        """
        AppLogger.__logger.debug(msg=message)

    @staticmethod
    def info(message: str) -> None:
        """Log at info or information level.

        :param message: Message to log
        :return: Nothing
        """
        AppLogger.__logger.info(msg=message)

    @staticmethod
    def warning(message: str) -> None:
        """Log at warning level.

        :param message: Message to log
        :return: Nothing
        """
        AppLogger.__logger.warning(msg=message)

    @staticmethod
    def error(message: str) -> None:
        """Log at error level.

        :param message: Message to log
        :return: Nothing
        """
        AppLogger.__logger.error(msg=message)

    @staticmethod
    def critical(message: str) -> None:
        """Log at critical level.

        :param message: Message to log
        :return: Nothing
        """
        AppLogger.__logger.critical(msg=message)

    @staticmethod
    def fatal(message: str) -> None:
        """Log at fatal level.

        :param message: Message to log
        :return: Nothing
        """
        AppLogger.__logger.fatal(msg=message)



"""Module containing the actual hardcoded logging json configs methods."""

# Custom libs
from zeppos_logging.app_logger_config_name import AppLoggerConfigName
from zeppos_logging.app_logger_config import AppLoggerConfig


# pylint: disable=too-few-public-methods
class AppLoggerConfigs(dict):
    """Class to define the actual json logging configs method."""

    def __init__(self) -> None:
        """Initialize the actual logging json config methods available.

        Note:
            Currently available configs are:
            - default_format_1
            - mqtt_format_1
        :return: Nothing
        """
        self[AppLoggerConfigName.DEFAULT_FORMAT_1] = AppLoggerConfig.default_format_1



"""Testing no module."""

# Standard libs
from unittest import TestCase


class TestNoTests(TestCase):
    """Unittest class to test the module."""

    def test_no_tests(self) -> None:
        """Test method: No tests here."""
        actual = 1
        result = 1
        self.assertEqual(result, actual)
